# pylama:ignore=W
"""
Payton main geometry module

Geometry module holds the basic geometry shapes. They are all inherited
from `payton.scene.geometry.base.Object` class. They are as simple as possible.

Their face informations are generated at the initialization but vertex array
object is not generated until it arrives in render pipeline.

* Object
  * Material (`payton.scene.material`)
    * Shader (`payton.scene.shader`)
"""
from payton.scene.geometry.base import Line, Object
from payton.scene.geometry.cube import Cube
from payton.scene.geometry.cylinder import Cylinder
from payton.scene.geometry.md2 import MD2
from payton.scene.geometry.mesh import Mesh
from payton.scene.geometry.plane import MatrixPlane, Plane
from payton.scene.geometry.point import PointCloud
from payton.scene.geometry.ragdoll import Joint, RagDoll
from payton.scene.geometry.sphere import Sphere
from payton.scene.geometry.wavefront import Wavefront

__all__ = [
    "Object",
    "Line",
    "Mesh",
    "Cylinder",
    "Cube",
    "Plane",
    "MatrixPlane",
    "Sphere",
    "PointCloud",
    "MD2",
    "Wavefront",
    "RagDoll",
    "Joint",
]



# pylama:ignore=W
"""#Payton Scene Module

##Main Scene Module:

* SDL2 Window
  * Scene (`payton.scene.scene.Scene`)
    * Geometry (`payton.scene.geometry`)
    * Grid (`payton.scene.grid`)
    * Background (`payton.scene.scene.Background`)
    * Clock (`payton.scene.clock`)
    * Lighting (`payton.scene.light`)
    * Collision Test (`payton.scene.collision`)
    * Observer (`payton.scene.observer`)
    * GUI (`payton.scene.gui`)
    * Material (`payton.scene.material`)
      * Shader (`payton.scene.shader`)
    * Type Support (`payton.scene.types`)

Also, `payton.scene.scene` contents are accessible through `payton.scene` main
module
"""
from payton.scene.scene import (
    SHADOW_HIGH,
    SHADOW_LOW,
    SHADOW_MID,
    SHADOW_NONE,
    Background,
    Scene,
)

__all__ = [
    "Scene",
    "Background",
    "SHADOW_HIGH",
    "SHADOW_LOW",
    "SHADOW_MID",
    "SHADOW_NONE",
]



# pylama:ignore=C901
import json
import logging
from typing import Any, Dict, Optional, cast

from payton.math.vector import invert_vector, plane_normal, vector_angle
from payton.scene.geometry.base import Object
from payton.scene.material import DEFAULT, Material
from payton.scene.types import VList


class Mesh(Object):
    """Mesh Object

    Mesh is almost like the Object except with some extra methods to make
    things easier. If you want to have custom geometries/shapes, it is
    better to extend `payton.scene.geometry.mesh.Mesh` instead of
    `payton.scene.geometry.base.Object`. Because Mesh will give you better
    and easier constructing capabilities such as adding triangles on the fly
    or sub-division or cutting and so forth. It is a way of designing objects
    by code.


    Example use case:

        .. include:: ../../../examples/basics/09_mesh.py
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    def clear_triangles(self) -> None:
        """Clear all triangles inside the Mesh"""
        self._vertices = []
        self._indices = []
        self._normals = []
        self._texcoords = []
        self._vertex_colors = []
        for material in self.materials.values():
            material._indices = []

        self.refresh()

    def to_dict(self) -> Dict[str, Any]:
        return {
            "vertices": self._vertices,
            "normals": self._normals,
            "texcoords": self._texcoords,
            "matrix": self.matrix,
            "materials": {
                mat: self.materials[mat].to_dict() for mat in self.materials
            },
            "children": {
                n: cast(Mesh, self.children[n]).to_dict()
                for n in self.children
                if isinstance(self.children[n], Mesh)
            },
        }

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> "Mesh":
        res = cls()
        res._vertices = d["vertices"]
        res._normals = d["normals"]
        res._texcoords = d["texcoords"]
        res.matrix = d["matrix"]
        res.materials = {
            n: Material.from_dict(d["materials"][n]) for n in d["materials"]
        }

        res.children = {
            n: cls.from_dict(d["children"][n]) for n in d["children"]
        }

        return res

    def to_json(self, **kwargs) -> str:
        return json.dumps(self.to_dict(), **kwargs)

    @classmethod
    def from_json(cls, jstr: str) -> "Mesh":
        data = json.loads(jstr)
        return cls.from_dict(data)

    def _calc_normal(self, i1: int, i2: int, i3: int, reverse: bool) -> bool:
        v1, v2, v3 = (
            self._vertices[i1],
            self._vertices[i2],
            self._vertices[i3],
        )
        normal = plane_normal(v1, v2, v3)
        if reverse:
            normal = invert_vector(normal)
        self._normals[i1] = normal
        self._normals[i2] = normal
        self._normals[i3] = normal
        return True

    def fix_normals(self, reverse=False) -> None:
        """Try to re-calculate Mesh normals, if your object has already perfect
        normals, do not call this method"""
        self._normals = [[0, 0, 1.0]] * len(self._vertices)

        [
            self._calc_normal(face[0], face[1], face[2], reverse)
            for face in self._indices
        ]

    def fix_texcoords(self, u: int = 1, v: int = 1) -> None:
        """Try to recalculate mesh texture coordinates by cube projection
        """
        self._texcoords = []
        self._calc_bounds()
        bbox = self.bounding_box
        vmin, vmax = bbox[0], bbox[1]
        width = vmax[0] - vmin[0]
        depth = vmax[1] - vmin[1]
        height = vmax[2] - vmin[2]
        normals = [
            [0.0, -1.0, 0.0],  # front
            [1.0, 0.0, 0.0],  # right
            [-1.0, 0.0, 0.0],  # left
            [0.0, 1.0, 0.0],  # back
            [0.0, 0.0, 1.0],  # top
            [0.0, 0.0, -1.0],  # bottom
        ]
        for face in self._indices:
            v1, v2, v3 = (
                self._vertices[face[0]],
                self._vertices[face[1]],
                self._vertices[face[2]],
            )
            # find face normal
            normal = plane_normal(v1, v2, v3)
            # determine which face of the cube
            angles = [vector_angle(normal, n) for n in normals]
            face_dir = angles.index(min(angles))
            if face_dir == 0:
                t1 = (v1[0] - vmin[0]) / width
                s1 = (v1[2] - vmin[2]) / height
                t2 = (v2[0] - vmin[0]) / width
                s2 = (v2[2] - vmin[2]) / height
                t3 = (v3[0] - vmin[0]) / width
                s3 = (v3[2] - vmin[2]) / height
                self._texcoords.extend([[t1, s1], [t2, s2], [t3, s3]])
            if face_dir == 1:
                t1 = (v1[1] - vmin[1]) / depth
                s1 = (v1[2] - vmin[2]) / height
                t2 = (v2[1] - vmin[1]) / depth
                s2 = (v2[2] - vmin[2]) / height
                t3 = (v3[1] - vmin[1]) / depth
                s3 = (v3[2] - vmin[2]) / height
                self._texcoords.extend([[t1, s1], [t2, s2], [t3, s3]])
            if face_dir == 2:
                t1 = (vmax[1] - v1[1]) / depth
                s1 = (v1[2] - vmin[2]) / height
                t2 = (vmax[1] - v2[1]) / depth
                s2 = (v2[2] - vmin[2]) / height
                t3 = (vmax[1] - v3[1]) / depth
                s3 = (v3[2] - vmin[2]) / height
                self._texcoords.extend([[t1, s1], [t2, s2], [t3, s3]])
            if face_dir == 3:
                t1 = (vmax[0] - v1[0]) / width
                s1 = (v1[2] - vmin[2]) / height
                t2 = (vmax[0] - v2[0]) / width
                s2 = (v2[2] - vmin[2]) / height
                t3 = (vmax[0] - v3[0]) / width
                s3 = (v3[2] - vmin[2]) / height
                self._texcoords.extend([[t1, s1], [t2, s2], [t3, s3]])
            if face_dir == 4:
                t1 = (v1[0] - vmin[0]) / width
                s1 = (v1[1] - vmin[1]) / depth
                t2 = (v2[0] - vmin[0]) / width
                s2 = (v2[1] - vmin[1]) / depth
                t3 = (v3[0] - vmin[0]) / width
                s3 = (v3[1] - vmin[1]) / depth
                self._texcoords.extend([[t1, s1], [t2, s2], [t3, s3]])
            if face_dir == 5:
                t1 = (v1[0] - vmin[0]) / width
                s1 = (v1[1] - vmin[1]) / depth
                t2 = (v2[0] - vmin[0]) / width
                s2 = (v2[1] - vmin[1]) / depth
                t3 = (v3[0] - vmin[0]) / width
                s3 = (v3[1] - vmin[1]) / depth
                self._texcoords.extend([[t1, s1], [t2, s2], [t3, s3]])
        self._texcoords = [[t[0] * u, t[1] * v] for t in self._texcoords]

    def add_triangle(
        self,
        vertices: VList,
        material: str = DEFAULT,
        normals: Optional[VList] = None,
        texcoords: Optional[VList] = None,
        colors: Optional[VList] = None,
    ) -> None:
        """Add triangle to Mesh

        Args:
          vertices: Vertices of the triangle. This is required. Ex:
                    `[[0, 0, 0], [2, 0, 0], [1, 1, 0]]`
          normals: Normals of the triangle. _(When left as None, Payton will
                   calculate the surface normal based on vertices and assign
                   it per given vertex.)_
          material: Name of the material to use.
          texcoords: Texture UV coordinates.
          colors: Per vertex colors (optional)
        """
        if len(vertices) != 3:
            logging.error("A triangle must have 3 vertices")
            return

        if normals is not None and len(normals) != 3:
            logging.error("There must be one normal per vertex")
            return

        if texcoords is not None and len(texcoords) != 3:
            logging.error("There must be one texcoord per vertex")
            return

        if normals is None:
            v1, v2, v3 = vertices[0], vertices[1], vertices[2]
            normal = plane_normal(v1, v2, v3)
            normals = [normal, normal, normal]
        if texcoords is None:
            texcoords = [[0, 0], [1, 0], [1, 1]]
        if colors:
            for color in colors:
                self._vertex_colors.append(color)

        self._vertices += vertices

        i = len(self._indices) * 3
        self._indices.append([i, i + 1, i + 2])
        self.materials[material]._indices.append([i, i + 1, i + 2])
        for normal in normals:
            self._normals.append(normal)
        for t in texcoords:
            self._texcoords.append(t)
        self.refresh()



from payton.scene.geometry.mesh import Mesh


def export_json(mesh: Mesh, filename: str, **kwargs):
    data = mesh.to_json(**kwargs)
    open(filename, "w").write(data)


def import_json(filename: str) -> Mesh:
    data = open(filename, "r").read()
    return Mesh.from_json(data)



"""Observer module
"""

import logging
import math
from typing import Any, List, Optional, Tuple, Type

import numpy as np  # type: ignore
import pyrr

from payton.math.geometry import raycast_plane_intersect
from payton.math.vector import invert_vector, sub_vector
from payton.scene.geometry.base import Object

BUTTON_LEFT = 1
BUTTON_RIGHT = 2


class Observer(object):
    """
    Observers are basically cameras in the scene.
    """

    def __init__(
        self,
        position: Optional[List[float]] = None,
        target: Optional[List[float]] = None,
        up: Optional[List[float]] = None,
        target_object: Optional[Type[Object]] = None,
        fov: float = 45.0,
        aspect_ratio: float = 1.33333,
        near: float = 0.1,
        far: float = 100.0,
        zoom: float = 10.0,
        active: bool = False,
        perspective: bool = True,
        **kwargs: Any,
    ) -> None:
        """
        Initialize the defaults of an Observer.
        Default observer is a perspective mode camera
        standing at x:10 y:10 z:10 looking at the origin
        and up direction is x:0 y:1 z:0

        Args:
          position: Position of the camera in absolute coordinates.
                    [10., 10., 10.]
          target: Target of the camera in absolute coordinates.
                  [0., 0., 0.]
          up: Up vector of the camera. [0., 0., 1.] Z-Up
          target_object: Instead of a fixed coordinate, track an object. (None)
          fov: Field of view of the camera [45.]
          aspect_ratio: Aspect ratio of the camera. [4/3]
          near: Near clipping plane, can't be negative. [0.1]
          far: Far plane. Further objects will be invisible. [100.]
          active: Is this the active camera in the scene? [False]
        """
        self.position: List[float] = [
            10.0,
            10.0,
            5.0,
        ] if position is None else position
        self.target: List[float] = [
            0.0,
            0.0,
            0.0,
        ] if target is None else target
        self.up: List[float] = [0.0, 0.0, 1.0] if up is None else up

        self.target_object: Optional[Type[Object]] = target_object
        self.fov: float = fov
        self.aspect_ratio: float = aspect_ratio
        self.near: float = near
        self.far: float = far

        # zoom factor for Orthographic projection
        self.zoom: float = zoom
        self.active: bool = active
        self.perspective: bool = perspective

        # Store matrices for future reference.
        self._projection: Optional[np.ndarray] = None
        self._view: Optional[np.ndarray] = None
        self._prev_intersection: Optional[np.ndarray] = None

    def distance(self) -> float:
        """
        Calculate distance to target
        """
        xdiff = self.position[0] - self.target[0]
        ydiff = self.position[1] - self.target[1]
        zdiff = self.position[2] - self.target[2]
        res = math.sqrt(xdiff * xdiff + ydiff * ydiff + zdiff * zdiff)
        if res == 0:
            res = 0.001
        return res

    def rotate_around_target(self, phi: float, theta: float) -> None:
        """
        Rotates camera around a subject.

        Args:
          theta: horizontal
          phi: vertical
        """
        diff = sub_vector(self.position, self.target)

        r = self.distance()
        _theta = math.acos(diff[2] / r)
        _phi = math.atan2(diff[1], diff[0])

        _theta = (math.degrees(_theta) + theta) % 360
        _phi = (math.degrees(_phi) + phi) % 360

        _theta = math.radians(_theta)
        _phi = math.radians(_phi)
        x = r * math.sin(_theta) * math.cos(_phi)
        y = r * math.sin(_theta) * math.sin(_phi)
        z = r * math.cos(_theta)

        self.position[0] = x + self.target[0]
        self.position[1] = y + self.target[1]
        self.position[2] = z + self.target[2]

    def pan(self, x: int, y: int, w: int, h: int) -> None:
        if self.target_object is not None:
            logging.error("Camera has an active target, can not pan")
            return
        px = math.ceil(w / 2)
        py = math.ceil(h / 2)
        _ceye, center_vector = self.screen_to_world(px, py, w, h)
        eye, vector = self.screen_to_world(x, y, w, h)
        center_vector = invert_vector(center_vector).copy() + [0.0]
        starget = self.target.copy() + [1.0]
        hit = raycast_plane_intersect(eye, vector, starget, center_vector)
        if hit is None:
            return
        if self._prev_intersection is None:
            self._prev_intersection = hit
        else:
            diff = np.subtract(self._prev_intersection, hit)
            self.position[0] += diff[0]
            self.position[1] += diff[1]
            self.position[2] += diff[2]

            self.target[0] += diff[0]
            self.target[1] += diff[1]
            self.target[2] += diff[2]

    def mouse(
        self,
        button: int,
        shift: bool,
        ctrl: bool,
        x: int,
        y: int,
        xrel: int,
        yrel: int,
        w: int,
        h: int,
    ) -> None:
        if shift and ctrl and button == BUTTON_LEFT:  # Panning
            self.pan(x, y, w, h)
            return

        if shift:
            if button == BUTTON_LEFT:
                self.rotate_around_target(xrel, -yrel)

        if ctrl:
            if button == BUTTON_LEFT:
                if self.perspective:
                    self.distance_to_target(self.distance() + yrel)
                else:
                    self.distance_to_target(self.zoom + yrel)

    def distance_to_target(self, distance: float) -> None:
        """Change distance to target

        This function does not change the spherical angles but just
        adjusts the distance to target coordinates
        """
        if not self.perspective:
            self.zoom = distance
            return
        diff = sub_vector(self.position, self.target)
        _theta = math.acos(diff[2] / self.distance())
        _phi = math.atan2(diff[1], diff[0])

        self.position[0] = self.target[0] + distance * math.sin(
            _theta
        ) * math.cos(_phi)
        self.position[1] = self.target[1] + distance * math.sin(
            _theta
        ) * math.sin(_phi)
        self.position[2] = self.target[2] + distance * math.cos(_theta)

    def render(self) -> Tuple[np.ndarray, np.ndarray]:
        """Render camera into two matrices.

        Return:
          (projection_matrix, view_matrix)
        """
        if self.perspective:
            proj_matrix = pyrr.matrix44.create_perspective_projection(
                self.fov,
                self.aspect_ratio,
                self.near,
                self.far,
                dtype=np.float32,
            )
        else:
            x = 70 * (self.aspect_ratio)
            y = 70
            if self.zoom == 0:
                self.zoom = 0.01
            if self.near < 1:
                self.near = 1
            proj_matrix = pyrr.matrix44.create_orthogonal_projection_matrix(
                left=(-x / self.zoom),
                right=(x / self.zoom),
                top=(y / self.zoom),
                bottom=(-y / self.zoom),
                near=self.near,
                far=self.far,
                dtype=np.float32,
            )

        self._projection = proj_matrix
        eye = np.array(self.position, dtype=np.float32)

        if self.target_object:
            # I believe there is a bug at mypy about @property methods
            self.target = self.target_object.position  # type: ignore

        target = np.array(self.target, dtype=np.float32)
        up = np.array(self.up, dtype=np.float32)
        view_matrix = pyrr.matrix44.create_look_at(eye, target, up)
        self._view = view_matrix
        return proj_matrix, view_matrix

    def screen_to_world(
        self, x: int, y: int, width: int, height: int
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Convert screen coordinates to world coordinates

        Unlike gluUnproject, this method returns a tuple of ray start
        (eye position) and ray vector (look at direction)

        Args:
          x: X Coordinate
          y: Y Coordinate
          width: Screen width
          height: Screen height

        Returns:
          (ray_start, ray_vector): Numpy arrays
        """
        x_f = (2.0 * x) / width - 1.0
        y_f = 1.0 - (2.0 * y) / height

        eye = np.array(
            [self.position[0], self.position[1], self.position[2], 1.0],
            dtype=np.float32,
        )

        ray_start = np.array([x_f, y_f, 1.0, 1.0], dtype=np.float32)
        proj = self._projection
        inv_proj = pyrr.matrix44.inverse(proj)
        eye_coords_ray = pyrr.matrix44.apply_to_vector(inv_proj, ray_start)
        eye_coords = np.array(
            [eye_coords_ray[0], eye_coords_ray[1], -1.0, 0.0], dtype=np.float32
        )
        view = self._view
        inv_view = pyrr.matrix44.inverse(view)
        ray_end = pyrr.matrix44.apply_to_vector(inv_view, eye_coords)

        if not self.perspective:
            ray_start = np.array([x_f, y_f, 1.0, 1.0], dtype=np.float32)
            eye_coords_ray = pyrr.matrix44.apply_to_vector(inv_proj, ray_start)
            eye_coords = np.array(
                [eye_coords_ray[0], eye_coords_ray[1], 0.0, 0.0],
                dtype=np.float32,
            )
            eye = pyrr.matrix44.apply_to_vector(inv_view, eye_coords)
            ray = ray_end - eye
            ray_dir = pyrr.vector.normalize(ray[0:4])
            eye[0] -= ray_dir[0] * 500
            eye[1] -= ray_dir[1] * 500
            eye[2] -= ray_dir[2] * 500
            eye[3] = 1

            return eye, ray_dir

        ray_dir = pyrr.vector.normalize(ray_end[0:4])
        return (eye, ray_dir)



"""Payton Shader

Payton Shader is an OpenGL 3.3 Compatible GLSL (330) module. Basic principle
is to be be able to compile vertex and fragment shaders and hold basic
ready-to-use shaders.

A usual user should not be aware of this shading system. Basic shaders defined
should be enough for simple simulations. Our intention is not to create a
graphics library. We want to create a simulation platform and let developers
to focus on mathematics rather than OpenGL Stuff.

Also, keep in mind that, this is NOT a generic-purpose shader class which can
be used at common OpenGL apps. This shader class is just enough of what is
essential to Payton Library. That is all.

"""

import ctypes
import logging
from typing import Any, Dict, List, Optional

import numpy as np  # type: ignore
from OpenGL.GL import (
    GL_FALSE,
    GL_FRAGMENT_SHADER,
    GL_GEOMETRY_SHADER,
    GL_PROGRAM_POINT_SIZE,
    GL_TRUE,
    GL_VERTEX_SHADER,
    glDisable,
    glEnable,
    glGetUniformLocation,
    glUniform1f,
    glUniform1i,
    glUniform3fv,
    glUniform4fv,
    glUniformMatrix4fv,
    glUseProgram,
    shaders,
)

depth_fragment_shader = """
#version 330 core
in vec4 FragPos;

uniform float far_plane;
uniform vec3 light_pos[100];

void main()
{
    float lightDistance = length(FragPos.xyz - light_pos[0]);
    lightDistance = lightDistance / far_plane;
    gl_FragDepth = lightDistance;
}
"""

depth_vertex_shader = """
#version 330 core
layout (location = 0) in vec3 aPos;

uniform mat4 model;

void main()
{
    gl_Position = model * vec4(aPos, 1.0);
}
"""

depth_geometry_shader = """
#version 330 core
layout (triangles) in;
layout (triangle_strip, max_vertices=18) out;

uniform mat4 shadowMatrices[6];

out vec4 FragPos; // FragPos from GS (output per emitvertex)

void main()
{
    for(int face = 0; face < 6; ++face)
    {
        gl_Layer = face;
        for(int i = 0; i < 3; ++i) // for each triangle's vertices
        {
            FragPos = gl_in[i].gl_Position;
            gl_Position = shadowMatrices[face] * FragPos;
            EmitVertex();
        }
        EndPrimitive();
    }
}
"""

default_fragment_shader = """
#version 330 core
out vec4 FragColor;

in vec2 tex_coords;
in vec3 frag_color;

in vec3 l_fragpos;
in vec3 l_normal;

uniform vec3 light_pos[100]; // assume 100 lights max.
uniform vec3 light_color[100];
uniform vec3 camera_pos;
uniform bool shadow_enabled;
uniform int samples;

uniform int LIGHT_COUNT;

uniform vec3 object_color;
uniform int material_mode;
uniform float opacity;
uniform float far_plane;
uniform int lit;

uniform sampler2D tex_unit;
uniform samplerCube depthMap;

// array of offset direction for sampling
vec3 gridSamplingDisk[20] = vec3[]
(
   vec3(1, 1,  1), vec3( 1, -1,  1), vec3(-1, -1,  1), vec3(-1, 1,  1),
   vec3(1, 1, -1), vec3( 1, -1, -1), vec3(-1, -1, -1), vec3(-1, 1, -1),
   vec3(1, 1,  0), vec3( 1, -1,  0), vec3(-1, -1,  0), vec3(-1, 1,  0),
   vec3(1, 0,  1), vec3(-1,  0,  1), vec3( 1,  0, -1), vec3(-1, 0, -1),
   vec3(0, 1,  1), vec3( 0, -1,  1), vec3( 0, -1, -1), vec3( 0, 1, -1)
);

float ShadowCalculation(vec3 fragPos)
{
    vec3 fragToLight = fragPos - light_pos[0];
    float currentDepth = length(fragToLight);
    float shadow = 0.0;
    float bias = 0.15;

    float viewDistance = length(camera_pos - fragPos);
    float diskRadius = (1.0 + (viewDistance / far_plane)) / 25.0;

    for(int i = 0; i < samples; ++i)
    {
        float closestDepth = texture(depthMap, fragToLight +
                                     gridSamplingDisk[i] * diskRadius).r;
        closestDepth *= 100.0;
        if(currentDepth - bias > closestDepth)
             shadow += 1.0;
    }

    return shadow / float(samples);
}

void main()
{
    vec3 color;
    float alpha;
    if (material_mode == 0 || material_mode == 2) {
        color = object_color;
        alpha = opacity;
    }
    if (material_mode == 1 || material_mode == 3) {
        color = texture(tex_unit, tex_coords).rgb;
        alpha = texture(tex_unit, tex_coords).a;
    }
    if (material_mode == 4) {
        color = frag_color;
        alpha = opacity;
    }

    if (lit == 0) {
        FragColor = vec4(color, alpha);
    } else {
        vec3 norm = normalize(l_normal);
        for (int i = 0; i < LIGHT_COUNT; i++) {
            vec3 ambient = 0.3 * light_color[i];

            // diffuse
            vec3 norm = normalize(l_normal);
            vec3 lightDir = normalize(light_pos[i] - l_fragpos);
            float diff = max(dot(norm, lightDir), 0.0);
            vec3 diffuse = diff * light_color[i];
            float spec = 0.0;
            vec3 halfwayDir = normalize(lightDir + camera_pos);
            spec = pow(max(dot(norm, halfwayDir), 0.0), 64.0);

            vec3 specular = spec * light_color[i];
            float shadow = shadow_enabled ? ShadowCalculation(l_fragpos) : 0.0;
            vec3 lighting = (ambient + (1.0 - shadow) *
                             (diffuse + specular)) * color;

            FragColor = vec4(lighting, alpha);
        }
    }
}"""  # type: str


default_vertex_shader = """
#version 330 core
layout ( location = 0 ) in vec3 position;
layout ( location = 1 ) in vec3 normal;
layout ( location = 2 ) in vec2 texCoords;
layout ( location = 3 ) in vec3 colors; // optional

out vec2 tex_coords;
out vec3 frag_color;

uniform mat4 model;
uniform mat4 view;
uniform mat4 projection;
uniform int view_mode;

out vec3 l_fragpos;
out vec3 l_normal;

void main()
{
    l_fragpos = vec3(model * vec4(position, 1.0));
    l_normal = mat3(transpose(inverse(model))) * normal;

    if (view_mode == 0) {
        gl_Position = projection * view * model * vec4(position, 1.0f);
    }
    if (view_mode == 1) {
        gl_Position = (projection * (vec4(position, 1.0f)
                        + vec4(model[3].xyz, 0.0f)));
    }

    gl_PointSize = 5.0;

    tex_coords = texCoords;
    frag_color = colors;
}
"""  # type: str


background_vertex_shader = """
#version 330 core
out vec2 v_uv;

void main()
{
  // uint idx = gl_VertexID;
  gl_Position = (vec4( gl_VertexID & 1, gl_VertexID >> 1, 0.0, 0.5 )
                 * 4.0 - 1.0);
  v_uv = vec2( gl_Position.xy * 0.5 + 0.5 );
}
"""  # type: str

background_fragment_shader = """
#version 330 core
uniform vec4 top_color;
uniform vec4 bot_color;
in vec2 v_uv;

out vec4 frag_color;
void main()
{
  frag_color = bot_color * (1 - v_uv.y) + top_color * v_uv.y;
}
"""  # type: str


class Shader(object):
    """Payton shader class. Creates shaders / programs

    For better performance, Shader aims to reduce the number of
    `glGetUniformlocation` function calls. For this purpose, `__init__` and
    `build` functions accept a list of variable names as `variables` argument.

    If variable not found in `self.variables` then the system will try to
    locate the variable location and store it in `self.variables` for
    future reference.

    This way, Shader goes through this list of variables after compiling the
    shader program and stores their locations in memory for future use.

    If you provide the list of variable names in build function call then
    it will overwrite the existing list.
    """

    NO_LIGHT_COLOR = 0  # type: int
    NO_LIGHT_TEXTURE = 1  # type: int
    LIGHT_COLOR = 2  # type: int
    LIGHT_TEXTURE = 3  # type: int
    PER_VERTEX_COLOR = 4  # type:int
    HUD = 5  # type: int
    DEPTH_CALC = 6  # type: int

    def __init__(
        self,
        fragment: str = default_fragment_shader,
        vertex: str = default_vertex_shader,
        geometry: str = "",
        variables: Optional[List[str]] = None,
        **kwargs: Any,
    ):
        """Initialize Shader.

        Args:
          fragment: Fragment shader code
          vertex: Vertex shader code
          variables: List of in/out/uniform variable names.
        """
        self.fragment_shader_source = fragment
        self.vertex_shader_source = vertex
        self.geometry_shader_source = geometry

        self.variables: List[str] = [] if variables is None else variables
        self._stack: Dict[str, int] = {}  # Variable stack.
        self._mode: int = self.NO_LIGHT_COLOR  # Lightless color material
        self._depth_shader = False

        self.program: int = -1

    def build(self) -> int:
        """Build GLSL Shader
        Compile shaders and compile glsl program
        Args:
          variables: List of in/out/uniform variable names.
        Return:
            self.program
        """
        vertex_shader = shaders.compileShader(
            self.vertex_shader_source, GL_VERTEX_SHADER
        )
        fragment_shader = shaders.compileShader(
            self.fragment_shader_source, GL_FRAGMENT_SHADER
        )
        geometry_shader = None
        if self.geometry_shader_source != "":
            geometry_shader = shaders.compileShader(
                self.geometry_shader_source, GL_GEOMETRY_SHADER
            )

        if geometry_shader is not None:
            self.program = shaders.compileProgram(
                vertex_shader, fragment_shader, geometry_shader
            )
        else:
            self.program = shaders.compileProgram(
                vertex_shader, fragment_shader
            )

        return self.program

    def use(self) -> bool:
        """Use GLSL Program.
        This method should be called before setting any variables or rendering
        any entities
        """
        if self.program == -1:
            logging.error("Shader not compiled")
            return False
        glUseProgram(self.program)
        glEnable(GL_PROGRAM_POINT_SIZE)
        return True

    def end(self) -> None:
        """Set the active GLSL program to 0
        """
        glUseProgram(0)
        glDisable(GL_PROGRAM_POINT_SIZE)

    def set_matrix4x4_np(
        self, variable: str, value: np.ndarray, transpose: bool = False
    ) -> bool:
        """Set 4x4 Numpy matrix value

        Some elements like Observer and Light can pass their matrices directly
        as numpy array to reduce number of object conversions. This is the
        ideal way if possible.

        Args:
          variable: Variable name to set
          value: Matrix to set. (Numpy matrix)
          transpose: Transpose matrix.
        """
        g_transpose = GL_TRUE if transpose else GL_FALSE
        location = self.get_location(variable)
        if location == -1:
            return False

        glUniformMatrix4fv(
            location,
            1,
            g_transpose,
            np.asfortranarray(value, dtype=np.float32),
        )
        return True

    def get_location(self, variable: str) -> int:
        """Get location of a given variable within the shader"""
        if variable in self._stack:
            return self._stack[variable]
        location = glGetUniformLocation(self.program, variable)

        self._stack[variable] = location
        return location

    def set_matrix4x4(
        self, variable: str, value: List[List[float]], transpose: bool = False
    ) -> None:
        """Set 4x4 Matrix value

        This method will simply convert `value` into numpy array and call
        `payton.scene.shader.Shader.set_matrix4x4_np` method.

        Args:
          variable: Variable name to set
          value: Matrix to set. (Numpy matrix)
          transpose: Transpose matrix.
        """
        self.set_matrix4x4_np(variable, np.array(value, np.float32), transpose)

    def set_vector3_array_np(
        self, variable: str, value: np.ndarray, count: int
    ) -> bool:
        """Set array of vec3 as numpy array value"""
        value = value.flatten()
        location = self.get_location(variable)
        if location == -1:
            return False
        glUniform3fv(location, count, value)
        return True

    def set_vector3_np(self, variable: str, value: np.ndarray) -> bool:
        """Set Vector 3 as numpy array value

        Some elements like Light or pre-set materials can pass their
        vertices directly as numpy array to reduce number of object
        conversions.

        Args:
          variable: Variable name to set
          value: Vector 3 to set. (Numpy array with 3 elemenets)
        """
        location = self.get_location(variable)
        if location == -1:
            return False
        glUniform3fv(location, 1, value)
        return True

    def set_vector4_np(self, variable: str, value: np.ndarray) -> bool:
        """Set Vector 4 as numpy array value

        Some elements like Light or pre-set materials can pass their vertices
        directly as numpy array to reduce number of object conversions.

        Args:
          variable: Variable name to set
          value: Vector 4 to set. (Numpy array with 4 elemenets)
        """
        location = self.get_location(variable)
        if location == -1:
            return False
        glUniform4fv(location, 1, value)
        return True

    def set_vector3(self, variable: str, value: List[float]) -> None:
        """Set Vector 3 as array value

        This method will simply convert `value` into numpy array and call
        `payton.scene.shader.Shader.set_vector3_np` method.

        Args:
          variable: Variable name to set
          value: Vector 3 to set. (Numpy array with 3 elemenets)
        """
        self.set_vector3_np(variable, np.array(value, dtype=np.float32))

    def set_int(self, variable: str, value: int) -> bool:
        """Set Integer value

        Args:
          variable: Variable name to set
          value: Integer value to set
        """
        location = self.get_location(variable)
        if location == -1:
            return False

        glUniform1i(location, ctypes.c_int(value))
        return True

    def set_float(self, variable: str, value: float) -> bool:
        """Set Float value

        Args:
          variable: Variable name to set
          value: Floaf value to set
        """
        location = self.get_location(variable)
        if location == -1:
            return False

        glUniform1f(location, ctypes.c_float(value))
        return True



import math
from typing import Any

from payton.scene.geometry.mesh import Mesh


class Cylinder(Mesh):
    """Cylinder Object

    Beware of using high values for number of meridians. You might end up with
    excessive number of vertices to render.

    Example use case:

        .. include:: ../../../examples/basics/18_cylinder.py

    """

    def __init__(
        self,
        bottom_radius: float = 0.5,
        top_radius: float = 0.5,
        meridians: int = 12,
        height: float = 1.0,
        **kwargs: Any,
    ) -> None:
        """Initialize the cylinder

        Args:
          bottom_radius: Radius at the bottom of the cylinder
          top_radius: Radius at the top of the cylinder
          meridians: Number of meridians/edges (as in geography)
          height: Height of the cylinder
        """
        super().__init__(**kwargs)
        self._bottom_radius: float = bottom_radius
        self._top_radius: float = top_radius
        self._meridians: int = meridians
        self._height: float = height
        self.build_cylinder()

    def build_cylinder(self) -> bool:
        step_angle = math.radians(360 / self._meridians)

        u_step = 1.0 / self._meridians
        r_bot = self._bottom_radius
        r_top = self._top_radius
        h_2 = self._height / 2.0
        for i in range(self._meridians):
            x1 = r_bot * math.cos(step_angle * i)
            y1 = r_bot * math.sin(step_angle * i)
            x2 = r_top * math.cos(step_angle * i)
            y2 = r_top * math.sin(step_angle * i)

            u1 = u_step * i
            v1 = 1.0
            u2 = u1
            v2 = 0.0

            x3 = r_bot * math.cos(step_angle * (i + 1))
            y3 = r_bot * math.sin(step_angle * (i + 1))
            x4 = r_top * math.cos(step_angle * (i + 1))
            y4 = r_top * math.sin(step_angle * (i + 1))

            u3 = u_step * (i + 1)
            v3 = 1.0
            u4 = u3
            v4 = 0.0

            self.add_triangle(
                [[x1, y1, -h_2], [x3, y3, -h_2], [x2, y2, h_2]],
                texcoords=[[u1, v1], [u3, v3], [u2, v2]],
            )

            self.add_triangle(
                [[x3, y3, -h_2], [x4, y4, h_2], [x2, y2, h_2]],
                texcoords=[[u3, v3], [u4, v4], [u2, v2]],
            )

            self.add_triangle(
                [[x1, y1, -h_2], [0.0, 0.0, -h_2], [x3, y3, -h_2]],
                texcoords=[[u1, v1], [0.0, 0.0], [u3, v3]],
            )

            self.add_triangle(
                [[x4, y4, h_2], [0.0, 0.0, h_2], [x2, y2, h_2]],
                texcoords=[[u4, v4], [0.0, 0.0], [u2, v2]],
            )

        self.fix_normals()
        return True



from itertools import product
from typing import Any, List

from payton.scene.geometry.mesh import Mesh
from payton.scene.material import SOLID, WHITE


class Plane(Mesh):
    """Plane object

    This is a 2D Plane in 3D World. Has a width in X and height in Y.
    If you need to place it in another axis, try modifying its matrix.

    Example use case:

        .. include:: ../../../examples/basics/13_plane.py
    """

    def __init__(
        self, width: float = 1.0, height: float = 1.0, **kwargs: Any
    ) -> None:
        """Initialize plane

        Args:
          width: Width of the plane
          height: Height of the plane
        """
        super().__init__(**kwargs)
        width *= 0.5
        height *= 0.5
        self._vertices = [
            [-width, -height, 0],
            [width, -height, 0],
            [width, height, 0],
            [-width, height, 0],
        ]
        self._normals = [[0, 0, 1], [0, 0, 1], [0, 0, 1], [0, 0, 1]]
        self._texcoords = [[-1, -1], [1, -1], [1, 1], [-1, 1]]
        self._indices = [[0, 1, 2], [2, 3, 0]]
        self.material._indices = self._indices


class MatrixPlane(Mesh):
    """Grid plane is a special plane which is generated by a NxM matrix and you
    can modify the z values by `.grid` list

    This is quite handy if you want to create a planar animation/simulation

    Example use case:

        .. include:: ../../../examples/mid-level/ripple.py
    """

    def __init__(
        self,
        width: float = 1.0,
        height: float = 1.0,
        x: int = 2,
        y: int = 2,
        **kwargs: Any,
    ) -> None:
        """Initialize MatrixPlane

        Args:
          x: Number of control points along X axis
          y: Number of control points along Y axis
          width: Total width
          height: Total height
        """
        super().__init__(**kwargs)
        self.width = width
        self.height = height
        self.x = x
        self.y = y
        if self.x < 2:
            self.x = 2
        if self.y < 2:
            self.y = 2
        self.grid: List[List[float]] = []
        # List of List for X, Y and List[float] for color.
        self.color_grid: List[List[List[float]]] = []
        self.populate_grid()

    def update_grid(self) -> None:
        # TODO This method needs some optimization
        for i, j in product(range(self.x), range(self.y)):
            self._vertices[(self.x * i) + j][2] = self.grid[i][j]
            self._vertex_colors[(self.x * i) + j] = self.color_grid[i][j]
        if self.material.display == SOLID:
            self.fix_normals()
        self.refresh()

    def populate_grid(self) -> None:
        """Turn grid data into plane"""
        self.clear_triangles()

        self.grid = [[0.0] * self.y for _i in range(self.x)]
        self.color_grid = [[WHITE] * self.y for _i in range(self.x)]

        step_x = self.width / (self.x - 1)
        step_y = self.height / (self.y - 1)
        step_u = 1.0 / (self.x - 1)
        step_v = 1.0 / (self.y - 1)

        c_x = self.width / 2.0
        c_y = self.height / 2.0
        for i, j in product(range(self.x), range(self.y)):
            self._vertices.append(
                [(i * step_x) - c_x, (j * step_y) - c_y, self.grid[i][j]]
            )
            self._texcoords.append([i * step_u, j * step_v])
            self._vertex_colors.append(WHITE)

            if i < self.x - 1 and j < self.y - 1:
                left = (self.x * i) + j
                right = (self.x * i) + (j + 1)
                top_left = (self.x * (i + 1)) + j
                top_right = (self.x * (i + 1)) + (j + 1)
                self._indices.append([top_right, right, left])
                self._indices.append([left, top_left, top_right])

        self.material._indices = self._indices
        self.fix_normals()
        self.refresh()



"""
What is a material?

Materials define how your scene entities look like. Their colors, shininess,
or displaying them as solid objects or wireframes, all are defined inside
object materials. This also effects if your object will respond to light
sources or not.

There are also predefined colors in this module.

* Material
  * Shader (`payton.scene.shader`)

"""
import os
from typing import Any, Dict, List, Optional

import numpy as np  # type: ignore
from OpenGL.GL import (
    GL_BLEND,
    GL_LINEAR,
    GL_LINEAR_MIPMAP_LINEAR,
    GL_ONE_MINUS_SRC_ALPHA,
    GL_REPEAT,
    GL_RGB,
    GL_RGBA,
    GL_SRC_ALPHA,
    GL_TEXTURE0,
    GL_TEXTURE_2D,
    GL_TEXTURE_MAG_FILTER,
    GL_TEXTURE_MIN_FILTER,
    GL_TEXTURE_WRAP_S,
    GL_TEXTURE_WRAP_T,
    GL_UNPACK_ALIGNMENT,
    GL_UNSIGNED_BYTE,
    glActiveTexture,
    glBindTexture,
    glBlendFunc,
    glEnable,
    glGenerateMipmap,
    glGenTextures,
    glPixelStorei,
    glTexImage2D,
    glTexParameterf,
)
from PIL import Image  # type: ignore

from payton.scene.shader import Shader
from payton.scene.types import IList

SOLID = 0  # type: int
WIREFRAME = 1  # type: int
POINTS = 2  # type: int

RED = [1.0, 0.0, 0.0]  # type: List[float]
GREEN = [0.0, 1.0, 0.0]  # type: List[float]
BLUE = [0.0, 0.0, 1.0]  # type: List[float]
CRIMSON = [220 / 255.0, 20 / 255.0, 60 / 255.0]  # type: List[float]
PINK = [1.0, 192 / 255.0, 203 / 255.0]  # type: List[float]
VIOLET_RED = [1.0, 62 / 255.0, 150 / 255.0]  # type: List[float]
DEEP_PINK = [1.0, 20 / 255.0, 147 / 255.0]  # type: List[float]
ORCHID = [218 / 255.0, 112 / 255.0, 214 / 255.0]  # type: List[float]
PURPLE = [128 / 255.0, 0.0, 128 / 255.0]  # type: List[float]
NAVY = [0.0, 0.0, 0.5]  # type: List[float]
ROYAL_BLUE = [65 / 255.0, 105 / 255.0, 225 / 255.0]  # type: List[float]
LIGHT_STEEL_BLUE = [176 / 255.0, 196 / 255.0, 222 / 255.0]  # type: List[float]
STEEL_BLUE = [70 / 255.0, 130 / 255.0, 180 / 255.0]  # type: List[float]
TURQUOISE = [0.0, 245 / 255.0, 1.0]  # type: List[float]
YELLOW = [1.0, 1.0, 0.0]  # type: List[float]
GOLD = [1.0, 225 / 255.0, 0.0]  # type: List[float]
ORANGE = [1.0, 165 / 255.0, 0.0]  # type: List[float]
WHITE = [1.0, 1.0, 1.0]  # type: List[float]
BLACK = [0.0, 0.0, 0.0]  # type: List[float]
DARK_GRAY = [0.2, 0.2, 0.2]  # type: List[float]
LIGHT_GRAY = [0.8, 0.8, 0.8]  # type: List[float]

DEFAULT = "default"
NO_VERTEX_ARRAY = -1
NO_INDICE = -2
EMPTY_VERTEX_ARRAY = -3


class Material(object):
    """
    Material information holder.
    """

    def __init__(
        self,
        color: Optional[List[float]] = None,
        display: int = SOLID,
        lights: bool = True,
        texture: str = "",
        opacity: float = 1.0,
        **kwargs: Any,
    ):
        """
        Initialize Material

        Color is constructed as a tuple of 3 floats. (Payton does not currently
        support transparency at MVP.) [1.0, 1.0, 1.0] are [Red, Green, Blue]

        Each element of color is a float between 0 and 1.
        (0 - 255 respectively)
        Also, there are predefined colors.

        Display Mode has 2 modes. Solid and Wireframe. Wireframe is
        often rendered in a faster way. Also good for debugging your
        object.

        Default variables:

            {'color': [1.0, 1.0, 1.0, 1.0],
             'display': SOLID}

        Args:
          color: Color of material
          display: Display type of material, SOLID / WIREFRAME (Default SOLID)
          lights: Effected by lights? (Default true)
          texture: Texture file name
          opacity: Opacity of the material (0 fully transparent, 1 opaque)
        """

        self.color: List[float] = [1.0, 1.0, 1.0] if color is None else color
        self.display: int = display
        self.lights: bool = lights
        self.texture: str = texture
        self.opacity: float = opacity
        self._image: Optional[Image] = None
        self._indices: IList = []
        self._vao: int = NO_VERTEX_ARRAY
        self._vbos: List[int] = []

        self._vertex_count: int = 0

        self._initialized: bool = False
        self._texture: Optional[int] = None

    def to_dict(self) -> Dict[str, Any]:
        return {
            "color": self.color,
            "display": self.display,
            "texture": self.texture,
            "opacity": self.opacity,
            "indices": self._indices,
        }

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> "Material":
        res = cls()
        res.color = d["color"]
        res.display = d["display"]
        res.texture = d["texture"]
        res.opacity = d["opacity"]
        res._indices = d["indices"]
        return res

    def build(self) -> bool:
        """Build material shaders

        Must be called at object build stage after generating vba.
        An active vba is required for building shader properly.
        """
        self._initialized = True
        if os.path.isfile(self.texture):
            img = Image.open(self.texture)
            self.load_texture(img)
        if self._image is not None:
            self.load_texture(self._image)
        return True

    def load_texture(self, img: Image) -> None:
        img_data = np.fromstring(img.tobytes(), np.uint8)
        width, height = img.size
        self._texture = glGenTextures(1)
        glPixelStorei(GL_UNPACK_ALIGNMENT, 1)
        glBindTexture(GL_TEXTURE_2D, self._texture)
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        glTexParameterf(
            GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR
        )
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT)
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT)

        mode = GL_RGBA
        if img.mode == "RGB":
            mode = GL_RGB
        if img.mode == "P":
            img = img.convert("RGB")
            img_data = np.fromstring(img.tobytes(), np.uint8)
            mode = GL_RGB
        glTexImage2D(
            GL_TEXTURE_2D,
            0,
            mode,
            width,
            height,
            0,
            mode,
            GL_UNSIGNED_BYTE,
            img_data,
        )
        glGenerateMipmap(GL_TEXTURE_2D)

    def refresh(self) -> None:
        self._initialized = False

    def render(
        self,
        model: np.ndarray,
        lit: bool,
        shader: Shader,
        mode: Optional[int] = None,
    ) -> None:
        """Render material

        This function must be called before rendering the actual object

        Args:
          model: Model matrix
          shader: Shader to be used for rendering
          lit: Is this a lit environment
          mode: Set explicit shader mode (optional - used for vertex colors)
        """
        if not self._initialized:
            self.build()

        _mode = Shader.LIGHT_COLOR

        if self.display == SOLID:
            if lit and self.lights:
                if self._texture is not None:
                    _mode = Shader.LIGHT_TEXTURE
                else:
                    _mode = Shader.LIGHT_COLOR
            else:
                if self._texture is not None:
                    _mode = Shader.NO_LIGHT_TEXTURE
                else:
                    _mode = Shader.NO_LIGHT_COLOR
        else:
            _mode = Shader.NO_LIGHT_COLOR

        if mode is not None:
            _mode = mode

        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

        shader.set_int("material_mode", _mode)
        shader.set_int("lit", 1 if lit else 0)
        if not self.lights:
            shader.set_int("lit", 0)

        shader.set_matrix4x4_np("model", model)
        shader.set_float("opacity", self.opacity)

        if self._texture is not None:
            check = shader.get_location("tex_unit")
            if check > -1:
                glActiveTexture(GL_TEXTURE0)
                glBindTexture(GL_TEXTURE_2D, self._texture)
                shader.set_int("tex_unit", 0)

        shader.set_vector3_np(
            "object_color", np.array(self.color, dtype=np.float32)
        )



"""Payton Line to Mesh module"""
from typing import List

from payton.math.matrix import create_rotation_matrix_raw
from payton.math.vector import add_vectors, scale_vector, vector_transform
from payton.scene.geometry import Line, Mesh


def extrude_line(line: Line, direction: List[float], distance: float) -> Mesh:
    """ Construct a mesh by extruding a line

    Args:
      line: Line to extrude
      direction: Direction vector to apply extrusion
      distance: Distance to extrude

    Returns:
      Mesh
    """
    vertices = line._vertices
    diff_vector = scale_vector(direction, distance)
    mirror_vertices = [add_vectors(v, diff_vector) for v in vertices]

    mesh = Mesh()
    for i in range(len(vertices) - 1):
        mesh.add_triangle(
            vertices=[vertices[i], vertices[i + 1], mirror_vertices[i]],
            texcoords=[[0.0, 0.0], [1.0, 0.0], [0.0, 1.0]],
        )
        mesh.add_triangle(
            vertices=[
                vertices[i + 1],
                mirror_vertices[i + 1],
                mirror_vertices[i],
            ],
            texcoords=[[1.0, 0.0], [1.0, 1.0], [0.0, 1.0]],
        )
    return mesh


def rotate_line(
    line: Line, axis: List[float], angle: float, steps: int = 10
) -> Mesh:
    """Create a mesh by rotating a line around axis at given angle (rad)

    Args:
      line: Line object to rotate
      axis: Axis to rotate line around (eg. [0, 0, 1.0])
      angle: Rotation angle
      steps: How many steps (resolution) to divide
    """
    step_angle = angle / steps
    vertices = line._vertices
    step_u = 1.0 / steps
    step_v = 1.0 / len(vertices)
    mesh = Mesh()

    for i in range(steps):
        matrix = create_rotation_matrix_raw(axis, step_angle)
        mirror_vertices = [vector_transform(v, matrix) for v in vertices]
        for j in range(len(vertices) - 1):
            mesh.add_triangle(
                vertices=[vertices[j], vertices[j + 1], mirror_vertices[j]],
                texcoords=[
                    [step_u * i, step_v * j],
                    [step_u * (i + 1), step_v * j],
                    [step_u * i, step_v * (j + 1)],
                ],
            )
            mesh.add_triangle(
                vertices=[
                    vertices[j + 1],
                    mirror_vertices[j + 1],
                    mirror_vertices[j],
                ],
                texcoords=[
                    [step_u * (i + 1), step_v * j],
                    [step_u * (i + 1), step_v * (j + 1)],
                    [step_u * i, step_v * (j + 1)],
                ],
            )
        vertices = mirror_vertices.copy()
    mesh.fix_normals()
    return mesh


def lines_to_mesh(lines: List[Line]) -> Mesh:
    """Fill the gap between line objects and construct a mesh

    Note: Algorithm follows the order of lines in the array, not their
          distances.

    Args:
      lines: List of line objects to use for construction.

    Returns:
      mesh
    """
    lens = [len(l._vertices) for l in lines]
    lmin = min(lens)
    lmax = max(lens)
    mesh = Mesh()
    if lmin != lmax:
        raise BaseException("Number of vertices for each line must be equal")
    if len(lines) == 0:
        raise BaseException("You must specify more than one line object")
    for i in range(len(lines) - 1):
        fvlist = lines[i]._vertices
        tvlist = lines[i + 1]._vertices
        for j in range(lmin - 1):
            mesh.add_triangle(vertices=[fvlist[j], fvlist[j + 1], tvlist[j]])
            mesh.add_triangle(
                vertices=[fvlist[j + 1], tvlist[j + 1], tvlist[j]]
            )
    mesh.fix_normals()
    mesh.fix_texcoords()
    return mesh



"""Ragdoll Object

Please note that, this is not a complete feature. Most likely many things
are still quite hard to accomplish using RagDoll. But it will be gradually
better in time. It just needs some time. Therefore, I am going to visit
this file every now and then.

Ragdoll is simply a basic skeleton.

Example use case:

    .. include:: ../../../examples/basics/28_ragdoll.py

"""
import math
from typing import Any, Dict

from payton.scene.geometry.base import Line
from payton.scene.geometry.cube import Cube
from payton.scene.geometry.sphere import Sphere
from payton.scene.material import WIREFRAME

R_COLLAR = "right_collar"
L_COLLAR = "left_collar"
R_SHOULDER = "right_shoulder"
L_SHOULDER = "left_shoulder"
NECK = "neck"
HEAD = "head"
R_HIP = "right_hip"
L_HIP = "left_hip"
R_ELBOW = "right_elbow"
L_ELBOW = "left_elbow"
R_KNEE = "right_knee"
L_KNEE = "left_knee"
R_WRIST = "right_wrist"
L_WRIST = "left_wrist"
R_FOOT = "right_foot"
L_FOOT = "left_foot"


class Joint(Sphere):
    """Joint object definition

    Joints are basically moving parts of a bone. For each
    bone, there is a starting (root) and end Joint.
    """

    def __init__(self, name: str, **kwargs):
        super().__init__(**kwargs)
        self.radius = 0.01
        self.material.display = WIREFRAME
        self.build_sphere()
        self.name = name


class Bone(Line):
    def __init__(
        self,
        root_joint_name: str,
        end_joint_name: str,
        length: float,
        **kwargs,
    ):
        super().__init__(**kwargs)
        root_joint = Joint(name=root_joint_name)
        end_joint = Joint(name=end_joint_name)
        self.add_child(root_joint_name, root_joint)
        self.root = self.children[root_joint_name]
        self.root.add_child(end_joint_name, end_joint)
        self.end = self.root.children[end_joint_name]
        self.end.position = [0.0, 0.0, length]

    def connect_to_root(self, joint: Joint):
        self.root.add_child(joint.name, joint)

    def connect_to_end(self, joint: Joint):
        self.end.add_child(joint.name, joint)


class RagDoll(Bone):
    def __init__(self, **kwargs: Any) -> None:
        kwargs["root_joint_name"] = "lower_spine"
        kwargs["end_joint_name"] = "higher_spine"
        kwargs["length"] = 1.0
        super().__init__(**kwargs)
        self.joints: Dict[str, Joint] = {}
        self.position = [0.0, 0.0, 2.0]
        self.setup_joints()
        self.setup_geometry()

    def setup_geometry(self):
        self.joints[HEAD].add_child(
            "mesh",
            Cube(from_corner=[-0.2, -0.2, 0], to_corner=[0.2, 0.2, 0.4]),
        )
        self.joints[NECK].add_child(
            "mesh",
            Cube(from_corner=[-0.1, -0.1, 0], to_corner=[0.1, 0.1, 0.2]),
        )
        self.joints[R_COLLAR].add_child(
            "mesh",
            Cube(from_corner=[-0.1, -0.1, 0], to_corner=[0.1, 0.1, 0.5]),
        )
        self.joints[L_COLLAR].add_child(
            "mesh",
            Cube(from_corner=[-0.1, -0.1, 0], to_corner=[0.1, 0.1, 0.5]),
        )
        self.joints[R_SHOULDER].add_child(
            "mesh",
            Cube(from_corner=[-0.1, -0.1, 0], to_corner=[0.1, 0.1, 0.5]),
        )
        self.joints[L_SHOULDER].add_child(
            "mesh",
            Cube(from_corner=[-0.1, -0.1, 0], to_corner=[0.1, 0.1, 0.5]),
        )
        self.joints[R_ELBOW].add_child(
            "mesh",
            Cube(from_corner=[-0.1, -0.1, 0], to_corner=[0.08, 0.08, 0.5]),
        )
        self.joints[L_ELBOW].add_child(
            "mesh",
            Cube(from_corner=[-0.08, -0.08, 0], to_corner=[0.08, 0.08, 0.5]),
        )
        self.joints[R_WRIST].add_child(
            "mesh",
            Cube(from_corner=[-0.1, -0.01, 0], to_corner=[0.1, 0.01, 0.2]),
        )
        self.joints[L_WRIST].add_child(
            "mesh",
            Cube(from_corner=[-0.1, -0.01, 0], to_corner=[0.1, 0.01, 0.2]),
        )
        self.root.add_child(
            "mesh",
            Cube(from_corner=[-0.3, -0.2, 0], to_corner=[0.3, 0.2, 1.0]),
        )
        self.joints[R_HIP].add_child(
            "mesh",
            Cube(from_corner=[-0.15, -0.15, 0], to_corner=[0.15, 0.15, 1.0]),
        )
        self.joints[L_HIP].add_child(
            "mesh",
            Cube(from_corner=[-0.15, -0.15, 0], to_corner=[0.15, 0.15, 1.0]),
        )
        self.joints[R_KNEE].add_child(
            "mesh",
            Cube(from_corner=[-0.1, -0.1, 0], to_corner=[0.1, 0.1, 1.0]),
        )
        self.joints[L_KNEE].add_child(
            "mesh",
            Cube(from_corner=[-0.1, -0.1, 0], to_corner=[0.1, 0.1, 1.0]),
        )
        self.joints[R_FOOT].add_child(
            "mesh",
            Cube(from_corner=[-0.05, -0.05, 0], to_corner=[0.05, 0.05, 0.3]),
        )
        self.joints[L_FOOT].add_child(
            "mesh",
            Cube(from_corner=[-0.08, -0.08, 0], to_corner=[0.08, 0.08, 0.3]),
        )

    def setup_joints(self):
        neck = Bone(
            root_joint_name="lower_neck", end_joint_name="head", length=0.2
        )
        self.connect_to_end(neck.root)
        head = Bone(
            root_joint_name="head", end_joint_name="head_top", length=0.4
        )
        neck.connect_to_end(head.root)
        left_collar = Bone(
            root_joint_name="l_collar_joint",
            end_joint_name="l_shoulder",
            length=0.5,
        )
        self.connect_to_end(left_collar.root)

        right_collar = Bone(
            root_joint_name="r_collar_joint",
            end_joint_name="r_shoulder",
            length=0.5,
        )
        self.connect_to_end(right_collar.root)

        right_collar.root.rotate_around_y(math.radians(100))
        left_collar.root.rotate_around_y(math.radians(-100))

        right_upper_arm = Bone(
            root_joint_name="r_shoulder", end_joint_name="r_elbow", length=0.5
        )
        right_collar.connect_to_end(right_upper_arm.root)
        right_upper_arm.rotate_around_y(math.radians(20))

        left_upper_arm = Bone(
            root_joint_name="l_shoulder", end_joint_name="l_elbow", length=0.5
        )
        left_collar.connect_to_end(left_upper_arm.root)
        left_upper_arm.rotate_around_y(math.radians(-20))

        right_lower_arm = Bone(
            root_joint_name="r_elbow", end_joint_name="r_wrist", length=0.5
        )
        right_upper_arm.connect_to_end(right_lower_arm.root)
        left_lower_arm = Bone(
            root_joint_name="l_elbow", end_joint_name="l_wrist", length=0.5
        )
        left_upper_arm.connect_to_end(left_lower_arm.root)

        right_hand = Bone(
            root_joint_name="r_wrist", end_joint_name="r_hand", length=0.2
        )
        right_lower_arm.connect_to_end(right_hand.root)
        left_hand = Bone(
            root_joint_name="l_wrist", end_joint_name="l_hand", length=0.2
        )
        left_lower_arm.connect_to_end(left_hand.root)

        right_upper_leg = Bone(
            root_joint_name="r_hip", end_joint_name="r_knee", length=1.0
        )
        self.connect_to_root(right_upper_leg.root)
        right_upper_leg.root.rotate_around_y(math.radians(160))

        left_upper_leg = Bone(
            root_joint_name="l_hip", end_joint_name="l_knee", length=1.0
        )
        self.connect_to_root(left_upper_leg.root)
        left_upper_leg.root.rotate_around_y(math.radians(-160))

        right_lower_leg = Bone(
            root_joint_name="r_knee", end_joint_name="r_foot", length=1.0
        )
        right_upper_leg.connect_to_end(right_lower_leg.root)
        left_lower_leg = Bone(
            root_joint_name="l_knee", end_joint_name="l_foot", length=1.0
        )
        left_upper_leg.connect_to_end(left_lower_leg.root)
        right_foot = Bone(
            root_joint_name="r_foot", end_joint_name="r_toe", length=0.3
        )
        right_lower_leg.connect_to_end(right_foot.root)
        right_foot.root.rotate_around_x(math.radians(90))

        left_foot = Bone(
            root_joint_name="l_foot", end_joint_name="l_toe", length=0.3
        )
        left_lower_leg.connect_to_end(left_foot.root)
        left_foot.root.rotate_around_x(math.radians(90))

        self.joints[NECK] = neck.root
        self.joints[HEAD] = head.root
        self.joints[R_COLLAR] = right_collar.root
        self.joints[L_COLLAR] = left_collar.root
        self.joints[R_SHOULDER] = right_upper_arm.root
        self.joints[L_SHOULDER] = left_upper_arm.root
        self.joints[R_ELBOW] = right_lower_arm.root
        self.joints[L_ELBOW] = left_lower_arm.root
        self.joints[R_WRIST] = right_hand.root
        self.joints[L_WRIST] = left_hand.root
        self.joints[R_HIP] = right_upper_leg.root
        self.joints[L_HIP] = left_upper_leg.root
        self.joints[R_KNEE] = right_lower_leg.root
        self.joints[L_KNEE] = left_lower_leg.root
        self.joints[R_FOOT] = right_foot.root
        self.joints[L_FOOT] = left_foot.root



"""
Clock is a special timer which runs in parallel to your scene and can do
many things.

Each clock has a period, scene and a callback function. Basically, for each
clock tick, the callback function gets called with `period` and `total`
variables.

- `period`: period of clock. (time difference between each click in seconds)
- `total`: total time past since the beginning of the simulation.

A clock can be used for several occasions like moving objects in time,
logging data, etc.

Each clock is technically a safe thread.

For accurate data calculations, time stops until the execution of the
callback function ends.

*Note:* All clocks are initially paused. You should hit *Space* to unpause
them. The main reason is, clock threads are created and started before
SDL2 and whole OpenGL scene is ready, so there is a chance that you can not
see the beginning of your time based simulations.

Example usage:

    .. include:: ../../examples/basics/04_clock.py
"""

import threading
import time
from typing import Any, Callable

# SAFE_ASSUMPTION constant is a wait time between each clock cycle.
SAFE_ASSUMPTION = 0.01


class ClockException(Exception):
    pass


class Clock(threading.Thread):
    """
    Each clock in Scene is actually a thread which has a
    reference of the scene. We carry this back reference as a Receiver
    Timer function carries a callback function to be called.
    """

    def __init__(
        self, period: float, callback: Callable[[float, float], None]
    ):
        """
        Args:
          name: Name of the clock (freetext for logging purposes)
          period: Clock cycle period of time between callbacks (seconds between
                  each tick)
          scene: Scene to be called back
          callback: Callback function to call when clock ticks.
        """
        threading.Thread.__init__(self)
        self.period = period
        self._total_time = 0.0
        self.callback = callback
        self._kill = False
        self._pause = True
        self._hold = False

    def kill(self) -> None:
        """
        Send a kill signal to existing clock.
        """
        self._kill = True

    def pause(self) -> None:
        """
        Send a pause signal to existing clock. Re-sending the same signal
        will let it continue.
        """
        self._pause = not self._pause

    def run(self) -> Any:
        """
        Callback function must have four arguments
        name of the clock for logging purposes
        scene reference to the actual scene object
        period of the clock
        total time difference from the initial run of the clock

        Return:
          bool
        """
        if self._kill:
            raise ClockException("This clock is already killed")
        while True:
            if self._kill:
                return False

            if self._pause:
                # sleep a safe time to avoid excessive cpu usage
                time.sleep(SAFE_ASSUMPTION)
                continue

            self.callback(self.period, self._total_time)

            time.sleep(self.period)
            self._total_time += self.period
        return True



# pylama:ignore=W
"""#Payton GUI Module
"""
from payton.scene.gui.base import Hud, Rectangle, Shape2D, Text
from payton.scene.gui.help import info_box
from payton.scene.gui.window import (
    Button,
    Panel,
    Theme,
    Window,
    WindowAlignment,
)

__all__ = [
    "Shape2D",
    "Rectangle",
    "Text",
    "Hud",
    "Window",
    "WindowAlignment",
    "Panel",
    "Button",
    "Theme",
    "info_box",
]



import typing

import numpy as np  # type: ignore
import pyrr

from payton.math.types import GArray

DIFF = 0.0000001


def point_project(p: GArray, origin: GArray, direction: GArray) -> float:
    return (
        direction[0] * (p[0] - origin[0])
        + direction[1] * (p[1] - origin[1])
        + direction[2] * (p[2] - origin[2])
    )


def distance(v1: np.ndarray, v2: np.ndarray) -> float:
    v3 = v2 - v1
    return pyrr.vector3.length(v3)


def distance2(v1: np.ndarray, v2: np.ndarray) -> float:
    v3 = v2 - v1
    return pyrr.vector3.length(v3) ** 2


def combine(v1: GArray, v2: GArray, f1: float, f2: float) -> np.ndarray:
    x = (f1 * v1[0]) + (f2 * v2[0])
    y = (f1 * v1[1]) + (f2 * v2[1])
    z = (f1 * v1[2]) + (f2 * v2[2])

    if len(v1) > 3 and len(v2) > 3:
        w = (f1 * v1[3]) + (f2 * v2[3])
        return np.array([x, y, z, w], dtype=np.float32)
    return np.array([x, y, z], dtype=np.float32)


def combine3(
    v1: GArray, v2: GArray, v3: GArray, f1: float, f2: float, f3: float
) -> np.ndarray:
    res = [
        (f1 * v1[0]) + (f2 * v2[0]) + (f3 * v3[0]),
        (f1 * v1[1]) + (f2 * v2[1]) + (f3 * v3[1]),
        (f1 * v1[2]) + (f2 * v2[2]) + (f3 * v3[2]),
        0,
    ]
    return np.array(res, dtype=np.float32)


def raycast_sphere_intersect(
    start: GArray, vector: GArray, sphere_center: GArray, sphere_radius: float
) -> bool:
    """Raycast Sphere Intersect Test

    Args:
      start: Start of the ray
      vector: Vector (direction) of the ray
      sphere_center: Sphere center in global coordinates
      sphere_radius: Sphere radius

    Returns:
      bool
    """
    proj = point_project(sphere_center, start, vector)
    if proj < 0:
        proj = 0.0
    vc = combine(start, vector, 1.0, proj)
    dist = distance2(sphere_center, vc)
    return dist < (sphere_radius ** 2)


def raycast_plane_intersect(
    start: GArray, vector: GArray, plane_point: GArray, plane_normal: GArray
) -> typing.Optional[np.ndarray]:
    """Raycast Plane Intersection Test

    Args:
      start: Start of the ray
      vector: Vector (direction) of the ray
      plane_point: Point on the plane
      plane_normal: Normal of the plane

    Returns:
      intersection: Intersection point or None
    """
    d = np.dot(vector, plane_normal)
    res = (d > DIFF) or (d < -DIFF)
    if not res:
        return None
    sp = np.subtract(plane_point, start)
    d = 1.0 / d
    t = np.dot(sp, plane_normal) * d
    if t <= 0:
        return None
    return combine(start, vector, 1.0, t)


def raycast_triangle_intersect(
    start: GArray, vector: GArray, p1: GArray, p2: GArray, p3: GArray
) -> typing.Tuple[
    typing.Union[np.ndarray, None], typing.Union[np.ndarray, None]
]:
    """Raycast Triangle Intersection Test

    Args:
      start: Start of the ray
      vector: Vector (direction) of the ray
      p1: P1 point of the triangle
      p2: P2 point of the triangle
      p3: P3 point of the triangle

    Returns:
      intersection, normal: Intersection point and intersection normal
      None, None: if there is no intersection
    """
    v1 = np.subtract(p2, p1)
    v2 = np.subtract(p3, p1)
    pvec = np.cross(vector, v2)
    det = np.dot(v1, pvec)
    if (det < DIFF) and (det > -DIFF):
        return None, None
    inv_det = 1.0 / det
    tvec = np.subtract(start, p1)
    u = np.dot(tvec, pvec) * inv_det
    if u < 0 or u > 1:
        return None, None
    qvec = np.cross(tvec, v1)
    v = np.dot(vector, qvec) * inv_det
    res = (v > 0) and (u + v <= 1.0)
    if not res:
        return None, None
    t = np.dot(v2, qvec) * inv_det
    if t <= 0:
        return None, None
    ip = combine(start, vector, 1, t)
    inor = np.cross(v1, v2)
    inor /= np.linalg.norm(inor)
    return ip, inor


def line_triangle_intersect(
    start: np.ndarray,
    end: np.ndarray,
    p1: np.ndarray,
    p2: np.ndarray,
    p3: np.ndarray,
) -> bool:
    """Check if line intersects a triangle in space

    Args:
      start: Staring position of the line
      end: End position of the line
      p1: Triangle corner coordinate 1
      p2: Triangle corner coordinate 2
      p3: Triangle corner coordinate 3
    """

    direction = end - start
    norm = np.linalg.norm(direction)
    if norm == 0:
        return False
    direction /= norm
    ip, _inor = raycast_triangle_intersect(start, direction, p1, p2, p3)
    if ip is None:
        return False
    # a---c----b
    d1 = distance(start, ip)
    d2 = distance(ip, end)
    d3 = distance(start, end)
    diff = d3 - (d1 + d2)
    return diff < DIFF and diff > -DIFF



# pylama:ignore=C901
# Wavefront Object File Support
# Only support ascii obj files and without material support.
# So pretty limited.
import logging
import os
from shutil import copyfile
from typing import Any, List

from payton.scene.geometry.mesh import Mesh
from payton.scene.material import DEFAULT, Material
from payton.scene.types import IList, VList


class Wavefront(Mesh):
    """
    Wavefront object file class.
    Only supports ascii obj files in a limited way.
    So do not depend so much on this class.
    Only designed to accept your triangular geometries.
    """

    def __init__(self, filename: str = "", **kwargs: Any) -> None:
        """
        Initialize Wavefront Object.
        """
        super().__init__()
        self.filename: str = filename
        self.path: str = ""
        if self.filename != "":
            self.load_file(self.filename)

    def load_file(self, filename: str) -> bool:
        """
        Load obj file.
        """
        if not os.path.isfile(filename):
            logging.exception(f"File not found {filename}")
            return False

        self.filename = filename
        self.path = os.path.dirname(os.path.abspath(self.filename))
        f = open(filename)
        data = f.read()
        f.close()
        self.load(data)
        return True

    def load_material(self, material_string: str):
        lines = material_string.splitlines()
        material = Material()
        material_name = DEFAULT
        for line in lines:
            parts = line.split(" ")

            if parts[0] == "newmtl":
                if material_name != DEFAULT:
                    self.add_material(material_name, material)
                material = Material()
                material_name = parts[1]

            if parts[0] == "Kd":
                # Currently we only support diffuse color
                material.color = [
                    float(parts[1]),
                    float(parts[2]),
                    float(parts[3]),
                ]

            if parts[0] == "map_Kd":
                rest = ' '.join(parts[1:])
                if rest == "/":
                    material.texture = rest
                else:
                    material.texture = f"{self.path}/{rest}"

        self.add_material(material_name, material)

    def load_material_file(self, filename: str) -> bool:
        if not os.path.isfile(filename):
            logging.exception(f"File not found {filename}")
            return False
        data = open(filename).read()
        return self.load_material(data)

    def load(self, obj_string: str) -> None:
        """
        A bit of information on file format,
        v -> x, y, z, (w)
        vt -> u, [v, (w)]
        vn -> x, y, z
        f -> vertex_index/texcoord_index/normal_index ...

        Also there are definitions of material and line and object name
        but for now, the assumption is there will always be triangulated
        wavefront object files and always a single object at a time.
        """
        _vertices: VList = []
        _indices: List[IList] = []
        _indice_materials: List[str] = []
        _normals: VList = []
        _texcoords: VList = []
        lines: List[str] = obj_string.splitlines()
        material = DEFAULT
        for line in lines:
            line = line.replace("  ", " ")
            parts = line.split(" ")
            command = parts[0].lower()
            if parts[0] == "mtllib":
                self.load_material_file(f"{self.path}/{parts[1]}")
            if command == "v":
                x = float(parts[1])
                y = float(parts[2])
                z = float(parts[3])
                _vertices.append([x, y, z])
            if command == "vt":
                u = float(parts[1])
                w = 1.0 - float(parts[2]) if len(parts) > 2 else 0
                _texcoords.append([u, w])
            if command == "vn":
                x = float(parts[1])
                y = float(parts[2])
                z = float(parts[3])
                _normals.append([x, y, z])
            if command == "usemtl":
                if parts[1] in self.materials:
                    material = parts[1]

            if command == "f":
                # I guess this part of the code should be compatable
                # with POLYGON as well but IDK.
                face = []  # type: List[List[int]]
                for i in range(len(parts)):
                    if parts[i] == "f" or parts[i] == "":
                        continue
                    subs = parts[i].split("/")
                    vertex = int(subs[0]) - 1
                    textcoord = (
                        int(subs[1]) - 1
                        if len(subs) > 1 and len(subs[1]) > 0
                        else -1
                    )
                    normal = (
                        int(subs[2]) - 1
                        if len(subs) > 2 and len(subs[2]) > 0
                        else -1
                    )
                    face.append([vertex, textcoord, normal])
                if len(face) > 3:
                    logging.error("Only triangular wavefronts are accepted")
                    return

                _indices.append(face)
                _indice_materials.append(material)

        # Now unpack indices to actual object data
        i = 0
        fix_normals = False

        for _k, index in enumerate(_indices):
            ind = []
            for f in index:
                l_vertex = _vertices[f[0]]
                if f[2] != -1:
                    l_normal = _normals[f[2]]
                else:
                    fix_normals = True
                    l_normal = [0.0, 0.0, 1.0]
                l_tex = [0.0, 0.0]
                if f[1] > -1:
                    l_tex = _texcoords[f[1]]
                self._vertices.append(l_vertex)
                self._normals.append(l_normal)
                self._texcoords.append(l_tex)
                ind.append(i)
                i += 1
            self._indices.append(ind)
            self.materials[_indice_materials[_k]]._indices.append(ind)

        if fix_normals:
            self.fix_normals()


def export(mesh: Mesh, filename: str, name: str = "object"):
    """Export mesh as wavefront object string
    @NOTE: Along with your object file, this function will create a separate
           .mtl file for the object material.

    Basic usage:

        from payton.scene.geometry import Cube
        from payton.scene.wavefront import export

        cube = Cube()
        export(cube, "cube.obj")

    Args:
      mesh: An instance of `payton.scene.geometry.mesh.Mesh`
      filename: Filename to write output file.
      name (optional): Name of the object, otherwise `object` will be used
    """
    if not isinstance(mesh, Mesh):
        logging.exception("Object is not an instance of Mesh")
        return None
    mat_filename = filename.replace(".obj", "") + ".mtl"

    output = [
        "# Payton Wavefront OBJ Exporter",
        f"mtllib {mat_filename}",
        f"o {name}",
    ]
    for v in mesh._vertices:
        output.append("v {}".format(" ".join([str(x) for x in v])))

    for t in mesh._texcoords:
        output.append("vt {}".format(" ".join([str(x) for x in t])))

    for n in mesh._normals:
        output.append("vn {}".format(" ".join([str(x) for x in n])))

    len_texcoords = len(mesh._texcoords) + 1
    len_normals = len(mesh._normals) + 1

    material_data = ["# Payton Wavefront OBJ Expoerter"]
    for name in mesh.materials:
        if len(mesh.materials[name]._indices) == 0:
            continue
        output.append(f"usemtl {name}")
        output.append("s off")
        material = mesh.materials[name]
        for indice in material._indices:
            f = [x + 1 for x in indice]
            t0 = str(f[0]) if len_texcoords > f[0] else ""
            n0 = str(f[0]) if len_normals > f[0] else ""

            t1 = str(f[1]) if len_texcoords > f[1] else ""
            n1 = str(f[1]) if len_normals > f[1] else ""

            t2 = str(f[2]) if len_texcoords > f[2] else ""
            n2 = str(f[2]) if len_normals > f[2] else ""
            output.append(
                f"f {f[0]}/{t0}/{n0} {f[1]}/{t1}/{n1} {f[2]}/{t2}/{n2}"
            )
        material_data.append(f"newmtl {name}")
        material_data.append(
            f"Kd {material.color[0]} {material.color[1]} {material.color[2]}"
        )
        if material.texture != "":
            base_name = os.path.basename(material.texture)
            path = os.path.join(
                os.path.abspath(os.path.dirname(filename)), base_name
            )
            copyfile(material.texture, path)
            material_data.append(f"map_Kd {base_name}")

    open(filename, "w").write("\n".join(output))
    open(mat_filename, "w").write("\n".join(material_data))



from payton.scene.gui.base import Hud
from payton.scene.gui.window import Button, Window, WindowAlignment


def help_win():
    text = """Payton Shortcuts and Mouse Controls:

MOUSE:
Zoom In-Out: Left CTRL + Mouse Drag (Up - Down)
Rotate: Left SHIFT + Mouse Drag (Left - Right)
KEYBOARD:
ESC: Quit
H: Show/Hide Help
C: Change Camera mode (Perspective/Orthographic)
Space: Start/Pause Scene Animations/Clocks
G: Show/Hide Grid
W: Display mode (wireframe/solid)
F2: Previous Observer F3: Next Observer
    """
    help_window = Window(
        align=WindowAlignment.RIGHT,
        title="How to use Payton Scene",
        height=100,
        width=400,
    )
    help_window.add_child(
        "text", Button(width=400, height=400, left=10, top=40, label=text)
    )
    return help_window


def info_box(left: int, top: int, width: int, height: int, label: str) -> Hud:
    hud = Hud()
    window = Window(
        "Info", width=width, height=height + 60, left=left, top=top
    )

    window.add_child(
        "label",
        Button(
            label=label, width=width - 20, height=height - 40, top=40, left=10
        ),
    )
    window.add_child(
        "close",
        Button(
            label="CLOSE",
            left=int(width / 2) - 10,
            top=height + 10,
            width=int(width / 2),
            height=40,
            on_click=window.hide,
        ),
    )
    hud.add_child("window", window)
    return hud



import unittest

import numpy as np  # type: ignore

from payton.math.geometry import (
    distance,
    distance2,
    point_project,
    raycast_triangle_intersect,
)


class TestGeometryMethods(unittest.TestCase):
    def test_raycast_triangle_intersect(self):
        start = np.array([0, 0, 5], dtype=np.float32)
        vector = np.array([0, 0, -1], dtype=np.float32)
        p1 = np.array([-1, -1, 0], dtype=np.float32)
        p2 = np.array([1, -1, 0], dtype=np.float32)
        p3 = np.array([0, 1, 0], dtype=np.float32)
        ip, inor = raycast_triangle_intersect(start, vector, p1, p2, p3)
        expected = np.array([0, 0, 0, 1], dtype=np.float32)
        expected_n = np.array([0, 0, 1], dtype=np.float32)
        np.testing.assert_array_almost_equal(ip, expected)
        np.testing.assert_array_almost_equal(inor, expected_n)

    def test_raycast_triangle_intersect_not(self):
        start = np.array([0, 0, 5], dtype=np.float32)
        vector = np.array([0, 0.5, 0.5], dtype=np.float32)
        p1 = np.array([-1, -1, 0], dtype=np.float32)
        p2 = np.array([1, -1, 0], dtype=np.float32)
        p3 = np.array([0, 1, 0], dtype=np.float32)
        ip, inor = raycast_triangle_intersect(start, vector, p1, p2, p3)
        self.assertIsNone(ip)
        self.assertIsNone(inor)

    def test_distance(self):
        p1 = np.array([2, 5, 3], dtype=np.float32)
        p2 = np.array([3, 9, -2], dtype=np.float32)
        d = distance(p1, p2)
        self.assertAlmostEqual(d, 6.4807405)

    def test_distance2(self):
        p1 = np.array([2, 5, 3], dtype=np.float32)
        p2 = np.array([3, 9, -2], dtype=np.float32)
        d = distance2(p1, p2)
        self.assertAlmostEqual(d, 41.999998039865204)

    def test_point_project(self):
        p = [2, 3, 1]
        origin = [4, 6, 9]
        direction = [0, 0, 1]
        expect = -8
        pp = point_project(p, origin, direction)
        self.assertEqual(pp, expect)


if __name__ == "__main__":
    unittest.main()



from enum import Enum
from typing import Any, Callable, Dict, Optional, Tuple

import numpy as np

from payton.scene.gui.base import Shape2D, Text
from payton.scene.shader import Shader


class Theme:
    """Main theme definitions for window elements
    """

    def __init__(self):
        self.opacity = 0.9

        self.background_color: Tuple[float, float, float] = [0.05, 0.05, 0.05]
        self.text_color: Tuple[float, float, float] = [1.0, 1.0, 1.0]
        self.title_background_color: Tuple[float, float, float] = [
            235 / 255,
            210 / 255,
            52 / 255,
        ]
        self.title_text_color: Tuple[float, float, float] = [0.0, 0.0, 0.0]


class WindowAlignment(Enum):
    """Window alignment

    Note that, in fixed cases (left, top, right, bottom) Width / Height
    information will be over-ridden.
    """

    FREE = "free"
    LEFT = "left"
    TOP = "top"
    RIGHT = "right"
    BOTTOM = "bottom"


class WindowElement(Shape2D):
    def __init__(
        self,
        width: int = 400,
        height: int = 300,
        left: int = 10,
        top: int = 10,
        align: WindowAlignment = WindowAlignment.FREE,
        theme: Optional[Theme] = None,
        **kwargs,
    ):
        kwargs["position"] = (left, top, 0)
        kwargs["size"] = (width, height)
        super().__init__(**kwargs)
        self._init: bool = False
        self.align = align
        self.theme = theme if theme is not None else Theme()
        self.material.opacity = self.theme.opacity

    def _reposition(self):
        if self.align == WindowAlignment.LEFT:
            self.position = [0.0, 0.0, 0.0]
            self.size = (self.size[0], self._parent_height)
        elif self.align == WindowAlignment.RIGHT:
            self.position = [self._parent_width - self.size[0], 0.0, 0.0]
            self.size = (self.size[0], self._parent_height)
        elif self.align == WindowAlignment.TOP:
            self.position = [0.0, 0.0, 0.0]
            self.size = (self._parent_width, self.size[1])
        elif self.align == WindowAlignment.BOTTOM:
            pos = [0, self._parent_height - self.size[1], 0]
            self.position = pos
            self.size = (self._parent_width, self.size[1])

    def draw(self):
        self._reposition()

    def render(
        self,
        lit: bool,
        shader: Shader,
        parent_matrix: Optional[np.ndarray] = None,
        _primitive: int = None,
    ) -> None:
        """Render the Text

        This calls the render method of
        `payton.scene.geometry.base.Object.render` then renders the text on top
        of the rectangle.
        """
        if not self._init:
            self.draw()

        super().render(lit, shader, parent_matrix)

    def add_child(self, name, obj: Shape2D) -> bool:  # type: ignore
        if obj.position[0] > self.size[0]:
            obj.position[0] = self.size[0] - 1
        total_v = obj.position[0] + obj.size[0]
        exceed = int(total_v - self.size[0])
        if exceed > 0:
            obj.size = (obj.size[0] - exceed, obj.size[1])

        if obj.position[1] > self.size[1]:
            obj.position[1] = self.size[1] - 1
        total_v = obj.position[1] + obj.size[1]
        exceed = int(total_v - self.size[0])
        if exceed > 0:
            obj.size = (obj.size[0], obj.size[1] - exceed)

        super().add_child(name, obj)
        obj.set_parent_size(self.size[0], self.size[1])


class Window(WindowElement):
    def __init__(
        self,
        title: str = "",
        width: int = 400,
        height: int = 300,
        left: int = 10,
        top: int = 10,
        align: WindowAlignment = WindowAlignment.FREE,
        theme: Optional[Theme] = None,
        **kwargs: Dict[str, Any],
    ):
        """Initialize Window

        @NOTE: I am aware that I could have just define "title" as other
               elemenets are already arguments of WindowElement but some
               code hinting and completion tools are a bit stupid. They
               can not inherit from parent class.
        """
        super().__init__(
            width=width,
            height=height,
            left=left,
            top=top,
            align=align,
            theme=theme,
            **kwargs,
        )
        self.title = title
        self.add_child(
            "title",
            Text(
                position=(0, 0, 0),
                size=(self.size[0], 20),
                label=self.title,
                # bgcolor=self.theme.title_background_color,
                color=self.theme.title_text_color,
                opacity=0.5,
            ),
        )

    def draw(self):
        super().draw()
        w, h = self.size[0], self.size[1]
        self.clear_triangles()
        if not self._init:
            self.add_triangle(
                [[0, 22, 1], [w, h, 1], [w, 22, 1]],
                texcoords=[[0, 0], [1, 1], [1, 0]],
                colors=[
                    self.theme.background_color,
                    self.theme.background_color,
                    self.theme.background_color,
                ],
            )
            self.add_triangle(
                [[0, 22, 1], [0, h, 1], [w, h, 1]],
                texcoords=[[0, 0], [0, 1], [1, 1]],
                colors=[
                    self.theme.background_color,
                    self.theme.background_color,
                    self.theme.background_color,
                ],
            )
            self.add_triangle(
                [[0, 0, 1], [w - 1, 22, 1], [w - 1, 0, 1]],
                texcoords=[[0, 0], [1, 1], [1, 0]],
                colors=[
                    self.theme.title_background_color,
                    self.theme.title_background_color,
                    self.theme.title_background_color,
                ],
            )
            self.add_triangle(
                [[0, 0, 1], [0, 22, 1], [w - 1, 22, 1]],
                texcoords=[[0, 0], [1, 1], [1, 0]],
                colors=[
                    self.theme.title_background_color,
                    self.theme.title_background_color,
                    self.theme.title_background_color,
                ],
            )
            self._init = True


class Panel(WindowElement):
    def draw(self):
        super().draw()
        w, h = self.size[0], self.size[1]
        self.clear_triangles()
        if not self._init:
            self.add_triangle(
                [[0, 0, 1], [w, h, 1], [w, 0, 1]],
                texcoords=[[0, 0], [1, 1], [1, 0]],
                colors=[
                    self.theme.title_background_color,
                    self.theme.title_background_color,
                    self.theme.title_background_color,
                ],
            )
            self.add_triangle(
                [[0, 0, 1], [0, h, 1], [w, h, 1]],
                texcoords=[[0, 0], [0, 1], [1, 1]],
                colors=[
                    self.theme.title_background_color,
                    self.theme.title_background_color,
                    self.theme.title_background_color,
                ],
            )
            self.add_triangle(
                [[1, 1, 1], [w - 1, h - 1, 1], [w - 1, 1, 1]],
                texcoords=[[0, 0], [1, 1], [1, 0]],
                colors=[
                    self.theme.background_color,
                    self.theme.background_color,
                    self.theme.background_color,
                ],
            )
            self.add_triangle(
                [[1, 1, 1], [1, h - 1, 1], [w - 1, h - 1, 1]],
                texcoords=[[0, 0], [0, 1], [1, 1]],
                colors=[
                    self.theme.background_color,
                    self.theme.background_color,
                    self.theme.background_color,
                ],
            )

            self._init = True


class Button(Panel):
    def __init__(
        self,
        label: str,
        width: int = 400,
        height: int = 300,
        left: int = 10,
        top: int = 10,
        align: WindowAlignment = WindowAlignment.FREE,
        theme: Optional[Theme] = None,
        on_click: Optional[Callable] = None,
        **kwargs: Any,
    ):
        kwargs["on_click"] = on_click
        super().__init__(
            width=width,
            height=height,
            left=left,
            top=top,
            align=align,
            theme=theme,
            **kwargs,
        )

        self._label = label
        self.text = Text(
            position=(0, 0, 1),
            size=(10, 10),
            label=label,
            color=self.theme.text_color,
        )
        self.add_child("label", self.text)

    def draw(self, **kwargs):
        super().draw()
        size = self.text.text_size
        x = (self.size[0] / 2) - (size[0] / 2)
        y = (self.size[1] / 2.0) - (size[1] / 2) - 4
        self.text.label = self._label
        self.text.position = [x, y]
        self.text.size = (size[0], size[1] + 4)

    @property
    def label(self) -> str:
        return self._label

    @label.setter
    def label(self, text: str):
        self._label = text
        self.text.label = self._label
